#ifdef FBGFX_FALLBACK
  ' missing symbols from freetype.bi
  function FT_Init_FreeType(p as any ptr) as integer
    return 0
  end function
  type FT_Face as integer
  /'
  end type '/
  type TFT_Var
    ErrorMsg  as integer
    Library   as integer
    PixelSize as integer
  end type
#elseif defined(__FB_PCOS__)
  ' missing symbols from freetype.bi
  function FT_Init_FreeType(p as any ptr) as integer
    return 0
  end function
  type FT_Face as LOGFONT ptr
  /'
  end type '/
  type TFT_Var
    ErrorMsg  as integer
    Library   as integer
    PixelSize as integer
  end type
  ' WinAPI for text output
  #include "windows.bi"
  #include "utf_conv.bi"
  #define FONTSIZE_FACTOR 1.2
  dim shared hWnd as HWND, hdc as HDC, vhdc as HDC, bmpi as BITMAPINFO
  hWnd = GetDesktopWindow()
  hdc = GetWindowDC(hWnd)
  vhdc = CreateCompatibleDC(hdc)
  setMapMode(hdc, MM_TEXT)
  bmpi = type(type<BITMAPINFOHEADER>( _
         sizeof(BITMAPINFOHEADER), SCREEN_X, -SCREEN_Y, 1, 32, BI_RGB, 0, 0, 0, 0, 0), {0, 0, 0, 0})
  sub destruct destructor
    DeleteDC(vhdc)
    ReleaseDC(hWnd, hdc)
  end sub
#else
  #include once "freetype2/freetype.bi"
  ' alpha blending
  const FT_MASK_RB_32 = &h00ff00ff, FT_MASK_AG_32 = &hff00ff00
  type TFT_Var
    ErrorMsg  as FT_Error
    Library   as FT_Library
    PixelSize as integer
  end type
#endif
#ifndef FONTSIZE_FACTOR
  #define FONTSIZE_FACTOR 1
#endif

type TCoord
  as integer x, y
end type

dim shared FT_Var as TFT_Var

declare sub drawGlyph(img as FB.Image ptr, byval font as FT_Face, byval x as integer, byval y as integer, byval clr as uinteger)
declare function FTsize(byval txt as string, byval font as FT_Face, byval size as integer = 18) as TCoord
declare function getFont(byval fontName as string) as FT_Face
declare sub centerFT overload(img as FB.Image ptr, txt as string, font as FT_Face, _
                     size as integer = 18, clr as uinteger = 0)
declare sub centerFT(x as integer, y as integer, txt as string, font as FT_Face, _
                     size as integer = 18, clr as uinteger = 0)
declare function printFT overload(byval txt as string, byval font as FT_Face, _
                         byval size as integer = 18, byval clr as uinteger = 0) as FB.Image ptr
declare function printFT(byval x as integer, byval y as integer, byval txt as string, byval font as FT_Face, _
                         byval size as integer = 18, byval clr as uinteger = 0) as boolean
declare function utf(txt as string) as string
declare function utf2ansi(txt as string) as string


sub drawGlyph(img as FB.Image ptr, byval font As FT_Face, byval x as integer, byval y as integer, byval clr as uinteger)
  #ifdef FBGFX_FALLBACK
  #elseif defined(__FB_PCOS__)
  #else
    if img = 0 then exit sub
    dim as FT_Bitmap BmpFT
    dim as ubyte ptr BmpPtr
    dim as ulong ptr dstPtr
    dim as integer   BmpHeight, BmpWidth, BmpPitch, imgX, imgPitch
    dim as uinteger  srcRB, srcAG, dstRB, dstAG, dstClr, alph
  
    imageinfo img, imgX,,, imgPitch
    imgPitch \= 4
    bmpFT     = font->Glyph->Bitmap
    bmpPtr    = bmpFT.Buffer
    bmpWidth  = bmpFT.Width
    bmpHeight = bmpFT.Rows
    bmpPitch  = imgX - BmpFT.Width
  
    dstPtr = cast(ulong ptr, img) + 8 + y*imgPitch + x
    do while bmpHeight
      do while bmpWidth
        srcRB    = clr and FT_MASK_RB_32
        srcAG    = clr and FT_MASK_AG_32
        dstClr   = *dstPtr
        alph     = *bmpPtr
        dstRB    = dstClr and FT_MASK_RB_32
        dstAG    = dstClr and FT_MASK_AG_32
        srcRB    = ((srcRB - dstRB) * alph) shr 8
        srcAG    = ((srcAG  - dstAG) shr 8)  * alph
        *dstPtr  = ((dstRB + srcRB) and FT_MASK_RB_32) or ((dstAG + srcAG) and FT_MASK_AG_32)
        dstPtr   += 1
        bmpPtr   += 1
        bmpWidth -= 1
      loop
      bmpWidth  = bmpFT.Width
      bmpHeight -= 1
      dstPtr += bmpPitch
    loop
  #endif
end sub

function FTsize(byval txt as string, byval font as FT_Face, byval size as integer = 18) as TCoord
  #ifdef FBGFX_FALLBACK
    return type<TCoord>(len(txt)*8, 16)
  #elseif defined(__FB_PCOS__)
    dim as HFONT hfont
    font->lfHeight = size
    hfont = createFontIndirect(font)
    selectObject(vhdc, hfont)
    dim as integer w = getTabbedTextExtentW(vhdc, txt, len(txt), 8, 0)
    deleteObject(hfont)
    return type<TCoord>(loword(w), hiword(w))
  #else
    dim as FT_Error ErrorMsg
    dim as FT_UInt GlyphIndex
    dim as FT_GlyphSlot slot = font->Glyph
    dim as integer px = 2, py = 0, tempInt
    if FT_Set_Pixel_Sizes(font, size, size) then return type(-1, -1)
    FT_Var.PixelSize = size
  
    for i as integer = 0 to len(txt)-1
      GlyphIndex = FT_Get_Char_Index(font, txt[i])
      if FT_Load_Glyph(font, GlyphIndex, FT_LOAD_DEFAULT) then return type(-1, -1)
      if FT_Render_Glyph(font->Glyph, FT_RENDER_MODE_NORMAL) then return type(-1, -1)
      px += slot->Advance.x shr 6
    next
    GlyphIndex = FT_Get_Char_Index(font, 124)
    if FT_Load_Glyph(font, GlyphIndex, FT_LOAD_DEFAULT) then return type(-1, -1)
    if FT_Render_Glyph(font->Glyph, FT_RENDER_MODE_NORMAL) then return type(-1, -1)
    py = font->Glyph->Bitmap_Top + font->Glyph->Bitmap.Rows
    return type(px, py)
  #endif
end function

function getFont(byval fontName as string) as FT_Face
  #ifdef FBGFX_FALLBACK
    return -1
  #elseif defined(__FB_PCOS__)
    dim face as FT_Face = new LOGFONT
    face->lfFaceName = "Arial"
    return face
  #else
    dim face as FT_Face
    dim ErrorMsg as FT_Error
    ErrorMsg = FT_New_Face(FT_Var.Library, fontName, 0, @face)
    if ErrorMsg then return 0
    return face
  #endif
end function

sub centerFT(img as FB.Image ptr, txt as string, font as FT_Face, size as integer = 18, clr as uinteger = 0)
  if img = 0 then exit sub
  dim as FB.Image ptr tempImg = printFT(txt, font, size, clr)
  dim as integer sx, sy, dx, dy, w, h
  if tempImg then
    if tempImg->width >= img->width then
      sx = 0
      dx = (tempImg->width-img->width)\2
      w  = img->width-1
    else
      sx = (img->width-tempImg->width)\2 + 1
      dx = 0
      w  = tempImg->width-1
    end if
    if tempImg->height-4 >= img->height then
      sy = 4
      dy = (tempImg->height-img->height)\2
      h  = img->height-1
    else
      sy = (img->height-tempImg->height)\2+4
      dy = 0
      h  = tempImg->height-1
    end if
    #ifdef FBGFX_FALLBACK
      put img, (sx, sy), tempImg, (dx, dy)-step(w, h)
    #else
      put img, (sx, sy), tempImg, (dx, dy)-step(w, h), alpha
    #endif
    imagedestroy tempImg
  end if
end sub

sub centerFT(x as integer, y as integer, txt as string, font as FT_Face, size as integer = 18, clr as uinteger = 0)
  dim as TCoord textsize = FTsize(txt, font, size)
  printFT x-textsize.x\2, y-textsize.y\2+2, txt, font, size, clr
end sub

function printFT(byval txt as string, byval font as FT_Face, _
                 byval size_ as integer = 18, byval clr as uinteger = 0) as FB.Image ptr
  dim as integer size = size_ * FONTSIZE_FACTOR
  dim as TCoord imgSize = FTsize(txt, font, size)
  if imgSize.x mod 4 then imgSize.x += 8 - (imgSize.x mod 4) else imgSize.x += 4
  dim as FB.Image ptr img = imagecreate(imgSize.x, imgSize.y, 0)
  #ifdef FBGFX_FALLBACK
    draw string img, (0, 0), txt, not clr
  #elseif defined(__FB_PCOS__)
    dim as HFONT hfont
    dim as HBITMAP bmp = CreateCompatibleBitmap(hdc, imgSize.x, imgSize.y)
    dim as BITMAPINFO bmpi = type( type<BITMAPINFOHEADER>(_
           sizeof(BITMAPINFOHEADER), imgSize.x, -imgSize.y, 1, 32, BI_RGB, 0, 0, 0, 0, 0), {0, 0, 0, 0})
    dim as uinteger ptr buf = new uinteger[imgSize.x*imgSize.y]
    dim as integer col, counter = 0

    font->lfHeight = size
    hfont = createFontIndirect(font)
    selectObject(vhdc, hfont)
    selectObject(vhdc, bmp)
    setBkMode(vhdc, OPAQUE)
    setTextColor(vhdc, clr or 1)
    textOut(vhdc, 0, 0, txt, len(txt))
    getDIBits(vhdc, bmp, 0, imgSize.y, buf, @bmpi, DIB_RGB_COLORS)

    for y as integer = 0 to imgSize.y-1
      for x as integer = 0 to imgSize.x-1
        col = buf[counter]
        if col then pset img, (x, y), col or &hff000000
        counter += 1
      next
    next
    deleteObject(hfont)
    deleteObject(bmp)
    delete[] buf
    'deallocate txtBuf
  #else
    dim as FT_UInt GlyphIndex
    dim as FT_GlyphSlot slot
    dim as integer px = 3, py = 0
    clr or= &hff000000
    if FT_Set_Pixel_Sizes(font, size, size) then imagedestroy img : return 0
    FT_Var.PixelSize = size
    slot = font->Glyph
  
    for i as integer = 0 to len(txt)-1
      GlyphIndex = FT_Get_Char_Index(font, txt[i])
      if FT_Load_Glyph(font, GlyphIndex, FT_LOAD_DEFAULT) then imagedestroy img : return 0
      if FT_Render_Glyph(font->Glyph, FT_RENDER_MODE_NORMAL) then imagedestroy img : return 0
  
      ' set pixels
      drawGlyph img, font, px + font->Glyph->Bitmap_Left, py - font->Glyph->Bitmap_Top + size, clr
      px += slot->Advance.x shr 6
    next
  #endif
  return img
end function

function printFT(byval x as integer, byval y as integer, byval txt as string, byval font as FT_Face, _
                 byval size as integer = 18, byval clr as uinteger = 0) as boolean
  #ifdef FBGFX_FALLBACK
    draw string(x, y), txt, clr
  #else
    dim as FB.Image ptr img = printFT(txt, font, size, clr)
    if img = 0 then return false
    put (x, y), img, alpha
    imagedestroy img
  #endif
  return true
end function

function utf(txt as string) as string
  #ifdef FBGFX_FALLBACK
    return utf2ansi(txt)
  #elseif defined(__FB_PCOS__)
    dim as wstring*999 txtBuf
    dim as integer txtSize = len(txt)
    UTFToWChar(UTF_ENCOD_UTF8, strptr(txt), @txtBuf, @txtSize)
    return txtBuf
  #else
    return wstr(txt)
  #endif
end function

function utf2ansi(txt as string) as string
  ' translate UTF8-string to ANSI codepage 437
  dim as string ret, tempString
  dim as integer i, index, char(128 to 255) = { _
      &hC7, &hFC, &hE9, &hE2, &hE4, &hE0, &hE5, &hE7, _
      &hEA, &hEB, &hE8, &hEF, &hEE, &hEC, &hC4, &hC5, _
      &hC9, &hE6, &hC6, &hF4, &hF6, &hF2, &hFB, &hF9, _
      &hFF, &hD6, &hDC, &hA2, &hA3, &hA5, &h20A7, &h192, _
      &hE1, &hED, &hF3, &hFA, &hF1, &hD1, &hAA, &hBA, _
      &hBF, &h2310, &hAC, &hBD, &hBC, &hA1, &hAB, &hBB, _
      &h2591, &h2592, &h2593, &h2502, &h2524, &h2561, &h2562, &h2556, _
      &h2555, &h2563, &h2551, &h2557, &h255D, &h255C, &h255B, &h2510, _
      &h2514, &h2534, &h252C, &h251C, &h2500, &h253C, &h255E, &h255F, _
      &h255A, &h2554, &h2569, &h2566, &h2560, &h2550, &h256C, &h2567, _
      &h2568, &h2564, &h2565, &h2559, &h2558, &h2552, &h2553, &h256B, _
      &h256A, &h2518, &h250C, &h2588, &h2584, &h258C, &h2590, &h2580, _
      &h3B1, &hDF, &h393, &h3C0, &h3A3, &h3C3, &hB5, &h3C4, _
      &h3A6, &h398, &h3A9, &h3B4, &h221E, &h3C6, &h3B5, &h2229, _
      &h2261, &hB1, &h2265, &h2264, &h2320, &h2321, &hF7, &h2248, _
      &hB0, &h2219, &hB7, &h221A, &h207F, &hB2, &h25A0, &hA0 }
  i = 0
  do while len(txt) > i
    index = txt[i]
    if index > 223 then
      if len(txt) > i+2 then
        index = ((index and 15) shl 12) + ((txt[i+1] and 63) shl 6) + (txt[i+2] and 63)
        i += 2
      end if
    elseif index > 191 then
      if len(txt) > i+1 then
        index = ((index and 31) shl 6) + (txt[i+1] and 63)
        i += 1
      end if
    end if
    select case index
      case 32 to 127
        ret &= mid(txt, i+1, 1)
      case else
        for k as integer = 128 to 255
          if index = char(k) then
            ret &= chr(k)
            exit for
          end if
        next
    end select
    i += 1
  loop
  return ret
end function
