const LANGUAGE = "spanish"
const MAX_WORDLENGTH = 30
const VERSION_MAJOR = 0, VERSION_MINOR = 4, VERSION_PATCH = 160212
const VERSION_STRING = VERSION_MAJOR & "." & VERSION_MINOR & "." & VERSION_PATCH
const SCREEN_X = 800, SCREEN_Y = 640
const B_COUNT = 6

enum RETVAR
  aborted = -99, windowClosed, generalError, emptyList
  ok = 0
end enum

#include once "vbcompat.bi"
#include once "fbgfx.bi"
#include once "include/font.bi"
dim shared font as FT_Face

declare function checkMouseIn(x as integer, y as integer, w as integer, h as integer, b as integer = -1) as boolean
declare function checkButtons(startX as integer, startY as integer, w as integer, h as integer, _
                   distX as integer, distY as integer, perLine as integer, _
                   count as integer, mouseButton as integer = -1) as integer
declare sub drawProgress(x as integer, y as integer, w as integer, h as integer, i as integer, all as integer)
declare sub flipCard(x as integer, y as integer, img1 as FB.Image ptr, img2 as FB.Image ptr = 0, _
                   bg as uinteger = &hffffff)
declare function loadDate(listname as string) as double
declare sub markButton(startX as integer, startY as integer, w as integer, h as integer, _
                   distX as integer, distY as integer, perLine as integer, _
                   no as integer, clr as uinteger = 0)
declare function message(msg as string, buttons as string = "OK", byval focus as integer = 0) as integer
declare function replace(txt as string, srch as string, repl as string) as string

#include once "include/TList.bi"
#include once "include/TExercise.bi"
#include once "include/TGui.bi"
#include once "include/help.bi"
#include once "include/levenshtein.bi"
#include once "include/multiput.bi"

declare sub quickSort(keyArray() as string, lpArray() as TListPoint, direction as integer = 1, _
                      byval s as integer = -1, byval e as integer = -1)
declare function randomWord overload(words as TList = TList) as TWord
declare function randomWord(words as TList, byval count as integer) as mdList(TWord)
declare function searchWord overload(word as string, pattern as string = "") as TWord
declare function searchWord(list as TList, word as string, pattern as string = "") as TWord

mdListDeclare(TExercise)
mdMapDeclare(long, long)
mdmapDeclare(string, long)
dim shared as mdList(TExercise)   globalExercises, globalOptions
dim shared as mdMap(long, long)   listOfLessions
dim shared as mdMap(string, long) listOfTags

' READ WORDS
dim as integer filenumber = freefile, search
open exepath & "/" & language & "/words.txt" for input as #filenumber
dim tempString as string, tempWord as TWord, tempLong as long
do until eof(filenumber)
  line input #filenumber, tempString
  select case lcase(left(tempString, 6))
    case "entry "
      if tempWord.id then globallist.put(tempWord.id, tempWord)
      tempWord.clear
      tempWord.id = cint(mid(tempString, 7))
      if globallist.containsKey(tempWord.id) then print "error: double entry " & tempWord.id
      tempWord.lession = tempWord.id \ 1000
      if listOfLessions.containsKey(tempWord.lession) then
        listOfLessions[tempWord.lession] += 1
      else
        listOfLessions.put(tempWord.lession, 1)
      end if
    case "word  "
      tempWord.word = trim(mid(tempString, 7))
    case "mark  "
      tempWord.mark = cint(mid(tempString, 7))
    case "group "
      tempWord.groupname = trim(mid(tempString, 7))
    case "trans "
      tempWord.translate.add(trim(mid(tempString, 7)))
    case else
      if lcase(left(tempString, 4)) = "tag " then
        tempString = trim(mid(tempString, 5))
        tempWord.tag.add(tempString)
        if listOfTags.containsKey(tempString) then
          listOfTags[tempString] += 1
        else
          listOfTags.put(tempString, 1)
        end if
      else
        search = instr(tempString, " ")
        tempWord.map.put(left(tempString, search-1), trim(mid(tempString, search+1)))
      end if
  end select
loop
if tempWord.id then globallist.put(tempWord.id, tempWord)
tempWord.clear
close #filenumber

' INITIALIZE FONT
FT_Var.ErrorMsg = FT_Init_FreeType(@FT_Var.Library)
if FT_Var.ErrorMsg then print "Could not load library"
font = getFont("font.ttf")

#include once "include/spanish/basic.bi"
if font = 0 then print "couldn't find font.ttf" : getkey : end
mainWindow

' END OF PROGRAM: SAVE CHANGES
kill exepath & "/" & language & "/words.bak"
name exepath & "/" & language & "/words.txt",  exepath & "/" & language & "/words.bak"
filenumber = freefile
open exepath & "/" & language & "/words.txt" for output as #filenumber
dim as mdSet(long) ks = globallist.keyset
foreach(long, tempLong, ks)
  print #filenumber, str(globallist.get(tempLong))
nexteach
close #filenumber

function checkButtons(startX as integer, startY as integer, w as integer, h as integer, _
                       distX as integer, distY as integer, perLine as integer, _
                       count as integer, b as integer = -1) as integer
  dim as integer mouseX, mouseY, mouseB
  if getmouse(mouseX, mouseY,, mouseB) <> 0 then return -1
  if b >= 0 andalso mouseB <> b then return -1
  for i as integer = 0 to count
    if checkMouseIn(startX+(i mod perLine)*(w+distX), startY+(i\perLine)*(h+distY), w, h) then return i
  next
  return -1
end function

function checkMouseIn(x as integer, y as integer, w as integer, h as integer, b as integer = -1) as boolean
  ' check whether mouse cursor is inside the given coordinates
  dim as integer mouseX, mouseY, mouseB
  if getmouse(mouseX, mouseY,, mouseB) <> 0 then return false
  if b >= 0 andalso mouseB <> b then return false
  return (mouseX > x and mouseY > y and mouseX < x+w and mouseY < y+h)
end function

sub drawProgress(x as integer, y as integer, w as integer, h as integer, i as integer, all as integer)
  ' draw progressing bar
  if i < 0 or i > all then exit sub
  line (x, y)-step(w, h), 0, b
  line (x+1, y+1)-step(w*i/all-2, h-2), COLOR_GRAY, bf
end sub

sub flipCard(x as integer, y as integer, img1 as FB.Image ptr, img2 as FB.Image ptr = 0, bg as uinteger = &hffffff)
  if img1 = 0 then exit sub
  for i as integer = 20 to 1 step -1
    screenlock
    line(x, y)-step(img1->width-1, img1->height-1), bg, bf
    multiPut , x+img1->width\2, y+img1->height\2, img1, i/20
    screenunlock
    sleep 10
  next
  line(x, y)-step(img1->width-1, img1->height-1), bg, bf
  if img2 = 0 then exit sub
  for i as integer = 1 to 20
    screenlock
    line(x, y)-step(img2->width-1, img2->height-1), bg, bf
    multiPut , x+img2->width\2, y+img2->height\2, img2, i/20
    screenunlock
    sleep 5
  next
end sub

function loadDate(listname as string) as double
  dim as integer filenumber = freefile
  dim as double ret
  open "/" & language & "/" & listname & ".lst" for binary as #filenumber
  get #filenumber,, ret
  close #filenumber
  return ret
end function

sub markButton(startX as integer, startY as integer, w as integer, h as integer, _
               distX as integer, distY as integer, perLine as integer, _
               no as integer, clr as uinteger = 0)
  dim as FB.Image ptr img = imagecreate(w-3, h-3, not clr)
  put (startX+(no mod perLine)*(w+distX)+2, startY+(no\perLine)*(h+distY)+2), img, xor
  imagedestroy(img)
end sub

function message(msg as string, buttons as string = "OK", byval focus as integer = 0) as integer
  dim as integer countLines = 1, countButtons = 1, srch, mx, my, mb
  dim as integer maxLines = 20, maxButtons = 10, fontSize = 14
  dim as integer msgHeight, msgWidth = 100, bWidth = 50    ' minimal width
  dim as integer maxWidth = SCREEN_X - 100
  dim as string lines(maxLines-1), butt(maxButtons-1), tempString, key
  dim as TGuiButton guiButton(1 to maxButtons)
  dim as TCoord coord
  dim as TGuiEvent event
  ' get message lines
  tempString = msg
  srch = instr(tempString, "|")
  do while srch
    lines(countLines-1) = left(tempString, srch-1)
    tempString = mid(tempString, srch+1)
    countLines += 1
    if countLines = maxLines then exit do
    srch = instr(tempString, "|")
  loop
  lines(countLines-1) = tempString
  ' get buttons
  tempString = buttons
  srch = instr(tempString, "|")
  do while srch
    butt(countButtons-1) = left(tempString, srch-1)
    tempString = mid(tempString, srch+1)
    countButtons += 1
    if countButtons = maxButtons then exit do
    srch = instr(tempString, "|")
  loop
  butt(countButtons-1) = tempString
  ' calculate geometry
  for i as integer = 0 to countLines-1
    coord = FTsize(lines(i), font, fontSize)
    if coord.x > msgWidth then msgWidth = coord.x
  next
  for i as integer = 0 to countButtons-1
    coord = FTsize(butt(i), font, fontSize)
    if coord.x > bWidth then bWidth = coord.x
  next
  msgHeight = coord.y*(countLines+2)
  ' extend box for buttons (if necessary)
  if msgWidth < (bWidth+30)*countButtons then msgWidth = (bWidth+30)*countButtons
  ' reduce box size if too big
  if msgWidth > maxWidth then msgWidth = maxWidth
  ' reduce buttons size if too big
  if msgWidth < (bWidth+30)*countButtons then bWidth = msgWidth\countButtons - 30
  ' draw box
  dim as FB.Image ptr lineImg, boxImg = imagecreate(msgWidth+40, msgHeight+40)
  dim as integer startX = (SCREEN_X-msgWidth)\2, startY = (SCREEN_Y-msgHeight)\2
  get (startX-20, startY-20)-step(msgWidth+39, msgHeight+39), boxImg
  line (startX-20, startY-20)-step(msgWidth+39, msgHeight+39), &hffffff, bf
  line (startX-19, startY-19)-step(msgWidth+37, msgHeight+37), 0, b
  line (startX-18, startY-18)-step(msgWidth+35, msgHeight+35), 0, b
  line (startX-16, startY-16)-step(msgWidth+31, msgHeight+31), 0, b
  ' show lines
  for i as integer = 0 to countLines-1
    lineImg = printFT(lines(i), font, fontSize)
    put (startX, startY+coord.y*i), lineImg, (0, 0)-step(msgWidth, coord.y), alpha
    imagedestroy lineImg
  next
  ' show buttons
  dim as TGui messageGui
  dim as integer bStartX = (SCREEN_X - (bWidth+30)*countButtons)\2 + 10
  for i as integer = 1 to countButtons
    guiButton(i) = TGuiButton(i, bStartX+(i-1)*(bWidth+30)-3, startY+msgHeight-10-coord.y, _
                              bWidth+15, coord.y+5, butt(i-1))
    if focus = i then guiButton(i).state or= GuiState.focus
    messageGui.buttons.add guiButton(i)
  next
  messageGui.show
  ' wait for user input
  do
    key = inkey
    select case key
      case chr(9)
        if focus > 0 and focus <= countButtons then
          guiButton(focus).state and= not GuiState.focus
          messageGui.setButton guiButton(focus)
          focus += 1
          if focus > countButtons then focus = 1
          guiButton(focus).state or= GuiState.focus
          messageGui.setButton guiButton(focus)
          messageGui.show
        end if
      case chr(13)
        if focus > 0 and focus <= countButtons then event.id = focus : exit do
    end select
    event = messageGui.getEvent
    if event.mButton > 0 and event.id > 0 then exit do
    sleep 1
  loop until key = chr(255, 107)
  ' clean up
  put (startX-20, startY-20), boxImg, pset
  imagedestroy boxImg
  if key = chr(255, 107) then return RETVAR.windowClosed
  do
    getmouse mx, my,, mb
    sleep 1
  loop until mb < 1
  return event.id
end function

sub quickSort(keyArray() as string, lpArray() as TListPoint, direction as integer = 1, _
         byval s as integer = -1, byval e as integer = -1)
  if s = -1 then s = lbound(keyArray)
  if e = -1 then e = ubound(keyArray)
  dim as integer i = s, j = e
  dim as string middle = keyArray((i+j)\2)
  do
    if direction > 0 then
      while keyArray(i) < middle : i += 1 : wend
      while keyArray(j) > middle : j -= 1 : wend
    else
      while keyArray(i) > middle : i += 1 : wend
      while keyArray(j) < middle : j -= 1 : wend
    end if
    if i <= j then
      swap keyArray(i), keyArray(j)
      swap lpArray(i), lpArray(j)
      i += 1
      j -= 1
    end if
  loop until i > j
  if j > s then quicksort(keyArray(), lpArray(), direction, s, j)
  if i < e then quicksort(keyArray(), lpArray(), direction, i, e)
end sub

function replace(txt as string, srch as string, repl as string) as string
  dim ret as string = txt, position as integer = instr(txt, srch)
  do while position
    ret = left(ret, position-1) & repl & mid(ret, position+len(srch))
    position = instr(ret, srch)
  loop
  return ret
end function

function randomWord(words as TList = TList) as TWord
  if words.list.isEmpty then return globallist.get(int(rnd*globallist.size))
  dim as mdList(TWord) list = randomWord(words, 1)
  if list.isEmpty then
    return TWord
  else
    return list.get(0)
  end if
end function

function randomWord(words as TList, byval count as integer) as mdList(TWord)
  dim w as TWord, ret as mdList(TWord), list as mdList(long), lp as TListPoint, tempInt as integer
  dim as integer wordSize = words.list.size
  foreach(TListPoint, lp, words.list)
    list.add(lp.id)
  nexteach
  if count > wordSize then count = wordSize
  for i as integer = wordSize to wordSize-count+1 step -1
    tempInt = int(rnd*i)
    w = globallist.get(list.get(tempInt))
    ret.add(w)
    list.remove(tempInt)
  next
  return ret
end function

function searchWord(word as string, padding as string = "") as TWord
  ' search the _first_ matching word in the global list
  dim w as TWord, i as long, ks as mdList(long) = globallist.keyset
  foreach(long, i, ks)
    w = globallist.get(i)
    select case padding
      case "trans"
        if w.translate.contains("word") then return w
      case "", "word"
        if w.word = word then return w
      case else
        if w.map.get(padding) = word then return w
    end select
  nexteach
  return TWord
end function

function searchWord(wordlist as TList, word as string, padding as string = "") as TWord
  ' search the _first_ matching word in a list
  dim l as TListPoint, w as TWord
  foreach(TListPoint, l, wordlist.list)
    w = globallist.get(l.id)
    select case padding
      case "trans"
        if w.translate.contains("word") then return w
      case "", "word"
        if w.word = word then return w
      case else
        if w.map.get(padding) = word then return w
    end select
  nexteach
  return TWord
end function
